/////////////////////////////////
//This code creates an 8-step CV sequencer with a Quantized Internal Oscillator, Speed Control,
//Gate Output, multiple Alternative Stepping Modes, a Run/Pause Switch, and a Reset Button
//
//!!! Create a change detection function so that knobs that have shift parameters only
//change value when they are moved. This will keep them from updating to a new knob position
//when the shift button is released.
//////////////////////////////////

//bool (boolean) data type stores a variable that only has two states: true or false
bool stepStop = false;   //Is it time to STOP the current step?
bool stepStart = false; //Is it time to START the next step?

//Set up an ARRAY to represent the step pins of the sequencer:
//every time the program sees the word "steps[n]" it will
//replace it with the "n" number in the list.
//e.g. from the array below: steps[0] = 9, steps[1] = 8, steps[6] = 3, etc
int steps[8] = {9, 8, 7, 6, 5, 4, 3, 2};  //the numbers represent the Arduino Digital Pins
int stepNum = 0; //variable to use as index for "steps[ ]"
int prevStepNum = 0; //variable to store the previous step number

//variables related to the tones
const int tonePin = 13;   //Output pin to produce tone
const int CVPin = A5;     //Analog Input Pin to read CV
int CVVal;      //Variable to store the current level of the CV Input (read on A5)
int note;           //frequency for the tone() function to play

//Variables for keeping track of stepping speed
const int speedKnobPin = A2;
const int durKnobPin = A3; //the duration knob pin is the same as the mode knob pin

//Variables for keeping track of the Mode
const int modeKnob = A3;  //center pin of Mode Knob is connected to pin A1
int modeRaw = 0;    //!!!variable to store the mode knob reading
int newMode = 1;    //!!!this will update the mode when the button is released
int mode = 1;       //!!!keep track of the current mode


//Variables for keeping track of the scale/root/octave
const int scaleKnob = A1; //Pin connection for scale knob
const int degreeKnob = A0;  //Pin connection for root knob


const int resetButtonPin = 10; //Pushbutton to reset the sequencer to the first step is connected to this pin
const int runsSwitchPin = 11; //Toggle switch to Run the sequencer is connected to this digital input
const int shiftButtonPin = 12; //Pushbutton to acces alternative modes of knobs
bool shift = HIGH;    //flag to keep track of whether or not the button is being pressed

const int gatePin = 0;    //pin on which to produce a digital gate out

void setup() {
  pinMode(steps[0], OUTPUT); //Configure pin 9 as an output
  pinMode(steps[1], OUTPUT); //Configure pin 8 as an output
  pinMode(steps[2], OUTPUT); //Configure pin 7 as an output
  pinMode(steps[3], OUTPUT);
  pinMode(steps[4], OUTPUT);
  pinMode(steps[5], OUTPUT);
  pinMode(steps[6], OUTPUT);
  pinMode(steps[7], OUTPUT);

  //INPUT_PULLUP mode is used for buttons and switches. Pressing the button will cause
  //the digitalRead() function to return "LOW"
  pinMode(resetButtonPin, INPUT_PULLUP); //configure pin as input pullup for reset button
  pinMode(runsSwitchPin, INPUT_PULLUP); //configure pin as input pullup
  pinMode(shiftButtonPin, INPUT_PULLUP); //configure button pin

  pinMode(gatePin, OUTPUT);

}

void loop() {
  //Much of the code goes inside a big if statement.
  //If the sequencer run switch is not "on" move to the "else" statement.
  if (digitalRead(runsSwitchPin) == LOW) { //if the switch is engaged, run the sequencer

    //stepController is a custom function that steps the variables startStep and stopStep
    //the two numbers represent 1. the speed of the steps and 2. the legth of the notes.
    //knobs are read into the funtion where they are scaled and used to control the sequencer
    //More details are in the body of the function declaration below (after the loop())
    stepController(speedKnobPin, durKnobPin); //add the duration knob here

    //Check to see if the stepController reports that it is time to STOP the current note
    if (stepStop == true) {
      noTone(tonePin);                    //stop tone
      digitalWrite(gatePin, LOW);         //toggle GATE LOW
      stepStop = false;                   //reset the flag
    }

    modeRaw = analogRead(modeKnob); //!!!Read Mode Knob
    
    //Check to see if the stepController reports that it is time to START the new note
    if (stepStart == true) {

// !!! This gets replaced by the changeRead logic
//      if (shift == HIGH) {  //only update mode when shift is not being pressed
//        //Use a knob to select from different operating modes.
//        //The map function is used to scale the number of modes down from 1024 to any desired number of modes
//        //These 4 modes (0-3) were chosen based on what seemed most useful and ordered in a sequence that seemed playable
//        mode = map(analogRead(modeKnob), 0, 1023, 0, 4); //read the knob attached to analog pin 1 and map to a range of 0 - 4
//      }

      if (changeRead(shiftButtonPin, modeKnob) == 2) { //!!!pass the shift pin and knob pin to changeRead 
        //if the knob has started moving since the shift button was released, update the mode
        //use a knob to select from different operating modes.
        //the map function is used to scale the number of modes down from 1024 to any desired number of modes
        //These 4 modes (0-3) were chosen based on what seemed most useful and ordered in a sequence that seemed playable
        newMode = map(modeRaw, 0, 1023, 0, 4); //read the knob attached to analog pin 1 and map to a range of 0 - 4
      }

      mode = newMode; //!!!update the mode if the knob has been intentionally changed, otherwise newMode will not have changed

      //A switch structure reads the variable "mode" and does the instructions in whatever case matches the current value of the variable
      switch (mode) {
        //Mode 0: HOLD
        case 0:
          //do nothing. This will just keep playing the same step.
          //Unless "reset" is pressed, then it steps forward by 1 each step
          if (digitalRead(resetButtonPin) == LOW) { //read the reset button. if pressed, step forward once each step
            stepNum++;
            if (stepNum > 7) {
              stepNum = 0;
            }
          }
          break;

        //Mode 1: STEP FORWARD
        case 1:
          if (stepNum == -1) { //check to see if reset was pressed while stopped (-1)
            stepNum = 0; //if reset while stopped, start over at 1st step
            break;
          }
          stepNum ++;  //increment stepNum up by one
          if (digitalRead(resetButtonPin) == LOW) { //check reset If pressed, step back to 1st step
            stepNum = 0;
          }
          //keep track of which note to play in the 8-note sequence. If we get to the end, start over at 0.
          if (stepNum > 7) {
            stepNum = 0;  //"0" is the frist note, "1" is the second, and so on
          }

          break;

        //Mode 2: STEP BACKWARDS
        case 2:

          stepNum --; //increment stepNum down by one
          //keep track of which note to play in the 8-note sequence. If we get to the beginning, start over at 7.
          if (stepNum < 0) {
            stepNum = 7;
          }
          if (digitalRead(resetButtonPin) == LOW) { //check reset
            stepNum = 7; //reset to 7 here, since we count backwards in this mode
          }
          break;

        //Mode 3: RANDOM STEPS
        case 3:
          if (digitalRead(resetButtonPin) == LOW) { //check reset
            stepNum = stepNum; //hold the current random step
          }
          else {  //if reset is not pressed, do the normal random behavior
            //use the random function to randomly pick a step between 0 and 7 each time through the loop
            stepNum = random(0, 8);
          }
          break;

        //Mode 4: STEP FORWARD BY TWOS
        case 4:
          stepNum += 2; //increment stepNum up by twos
          //keep track of which note to play in the 8-note sequence. If we get to the end, start over at 0.
          if (stepNum > 7) {
            stepNum = 0;  //"0" is the frist note, "1" is the second, and so on
          }
          if (digitalRead(resetButtonPin) == LOW) { //check reset
            stepNum = 0;
          }
          break;
      }

      digitalWrite(steps[prevStepNum], LOW); //turn off the last step

      digitalWrite(steps[stepNum], HIGH); //turn on the next step
      prevStepNum = stepNum;  //save this step so we can turn it off before the next new one.
      stepStart = false; //set our running state to indicate that we are not ready to start a new step

      //get value of CV from pin A5
      //Note: the (void) and the delay are there to deal with ghost voltages
      //resulting from multiplexing of the ADC and the unbuffered circuitry.
      //changes to the hardware can make this unnecessary
      (void)analogRead(CVPin);
      delay(8);
      CVVal = analogRead(CVPin);

      //Write the frequency to the tone pin to play through the audio out
      toneQuantize(CVVal);  //custom function reads the CV input pin, maps the frequency to a scale, and sets the varible "note"


      if (CVVal == 0) { //when the knob is all the way down, turn off the audio and gate output
        noTone(tonePin);
        digitalWrite(gatePin, LOW);
      }

      else { //when the knob is in any position other than zero, play the quanitized note
        tone(tonePin, note);  //for all other notes, write the quantized note to the audio out
        digitalWrite(gatePin, HIGH);  //turn the gate signal to +5V
      }
    }
  }

  else { //Do this if the Run Switch is not engaged
    noTone(tonePin);  //turn off any note if run switch disengaged

    if (digitalRead(resetButtonPin) == LOW) { //check reset even when not running
      stepNum = -1; //indicate reset while stopped
    }
  }
}


//CUSTOM STEPPING FUNCTION
void stepController(int speedIn, int durIn) {
  //Variables for Keeping Track of Time (unsigned long will allow us to store very large numbers)
  static unsigned long previousTime = 0; //keep track of the time we last stepped
  static unsigned long counter = 0;  //keep track of time since the last step
 
  counter = millis() - previousTime;    //counter keeps track of the number of ms since last step began
  
  int speedValRaw = analogRead(speedIn); //read the speed knob and store the raw value
  int stepInterval = map(speedValRaw, 0, 1023, 300, 10); //map the value to new speeds (chosen by ear)
  static int stepDuration = stepInterval - 1; //this will set the duration if the shift button has not been pressed
  static int durValRaw = 1023; //initalize value

// !!!This gets replaced by changeRead Logic
//  if (shift == LOW) { //check the shift button
//    //read the knob and update the variable only when the shift button is pressed
//    durValRaw = analogRead(durKnobPin);
//  }

  int durChangeVal = analogRead(durKnobPin); //!!!read the duration knob and store the raw value

    if (changeRead(shiftButtonPin, durKnobPin)  == 1) { //!!! check the shift button and the knob to see if they have changed
      durValRaw = durChangeVal;   //!!!if the knob has been moved since the button was pressed, store new value to durValRaw
    }

  //Update the length of the note to keep it relative to the speed
  stepDuration = map(durValRaw, 0, 1023, 0, stepInterval - 1);

  if (counter > stepDuration) { //check to see if we should STOP the previous step
    stepStop = true;    //set flag to trigger the stop of the current step
  }

  //Add an if statment to check for the knob to be at Zero. If so, stop the sequencer
  if (speedValRaw != 0) { //if knob is not set to 0, it sets the speed of the sequencer
    if (counter >= stepInterval && stepStart == false) { //check to see if we should START the next step
      stepStart = true;   //set flag to trigger the start of the next step
      previousTime = millis(); //set new previous time to restart counter at beginning of each step
    }

    else {
      stepStart = false;    //make it unambiguous that we should NOT start a new step
    }
  }
  else { //if knob is at zero, stop sequencer
    stepStart = false;
    stepStop = false;
  }

}

//NOTE QUANTIZER
void toneQuantize(int rawVal) {
  static int  scaleType = 0;
  //31 = B0, 33 = C1, 35 = Csharp1, 37 = D1, and so on ... 4978 = Dsharp8
  int chromaticScale[] = {31, 33, 35, 37, 39, 41, 44, 46, 49, 52, 55, 58, 62, 65, 69, 73, 78, 82, 87, 93, 98, 104, 110, 117, 123, 131, 139, 147, 156, 165, 175, 185, 196, 208, 220, 233, 247, 262, 277, 294, 311, 330, 349, 370, 392, 415, 440, 466, 494, 523, 554, 587, 622, 659, 698, 740, 784, 831, 880, 932, 988, 1047, 1109, 1175, 1245, 1319, 1397, 1480, 1568, 1661, 1760, 1865, 1976, 2093, 2217, 2349, 2489, 2637, 2794, 2960, 3136, 3322, 3520, 3729, 3951, 4186, 4435, 4699, 4978};
  //These masks are used to create a list of notes from the chomatic scale that conform to one octave of other scale types
  int chromaticMask[14] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13}; //the sequential numbers here just move through the choromatic scale
  int majorMask[8] = {0, 2, 4, 5, 7, 9, 11, 12}; //These numbers are used as index values to only read the notes from the chromatic scale that confrom to a major scale
  int minorMask[8] = {0, 2, 3, 5, 7, 8, 10, 12};
  int majorPentatonicMask[6] = {0, 2, 4, 7, 9, 12};
  int minorPentatonicMask[6] = {0, 3, 5, 7, 10, 12};

  int tonicMask[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11}; //add an offset to represent which note to start the scale on
  int octaveMask[] = {0, 12, 24, 36, 48, 60}; //add an offset to move up by octaves
  int pitchNum;   //variable to store the index of the note array

  static int quality;
  static int octave = 1; //can be made variable by adding a knob
  static int octaveOffset = 0;  //used for multiple-octave scale modes
  static int root = 1;  //can be made variable by adding a knob (0 = B, 1 = C, 2 = C#, 3 = D ... 6 = F, etc)
  int i;  //variable to represent index of array

  int scaleRaw = analogRead(scaleKnob); //!!!read the scale knob and store a raw value

  //!!! All of this gets replaced with changeReadTwo logic
  //  //check the shift button, if it is pressed, change the octave with the scale knob
  //  shift = digitalRead(shiftButtonPin);
  //  if (shift == LOW) {
  //    octave = map(analogRead(scaleKnob), 0, 1023, 0, 3);
  //  }
  //
  //  else { // if the shift button is not pressed, change the scale type with the scale knob as usual
  //    scaleType = analogRead(scaleKnob); //read the scale knob and store the value in "scaleType"
  //  }

  changeReadTwo(shiftButtonPin, scaleKnob); //!!!read the shift button and the shifted knob
  if (changeReadTwo(shiftButtonPin, scaleKnob) == 1) { //!!!the function will return a 1 if the knob has moved since button press
    octave = map(scaleRaw, 0, 1023, 0, 2);
  }

  else if (changeReadTwo(shiftButtonPin, scaleKnob) == 2) { //!!!the function will return a 2 if the knob has moved since button release
    scaleType = scaleRaw;
  }

  int degree = analogRead(degreeKnob); //store a raw value to set which step of the scale the knob will start on

  int scale = map(scaleType, 0, 1023, 0, 4); //map scale type to select from one of 5 scales

  switch (scale) {

    case 0:   //Chromatic
      //three octaves of a chromatic scale = (12*3) + 2 for a zero for silence and one extra note for a high octave root
      //i will map the rawVal reading from the variable that is passed to this function, presumably from a
      //10-bit knob reading
      i = map(rawVal, 0, 1023, 0, 38); //rawVal gets mapped to index variable "i"
      i = i + map(degree, 0, 1023, 0, 12);
      octaveOffset = octave + (i / 12); //divide i by 12, and stores anything left of the decimal + 1 into octaveOffset
      i = i % 12; //% means divide by 12 and store anything to the right of the decimal as i (this lets us count in base 12)
      quality = chromaticMask[i]; //choose the scale quality mask
      pitchNum = octaveMask[octaveOffset] + quality + tonicMask[root]; //pitchNum is our new quantized index value
      note = chromaticScale[pitchNum]; //pick our note from the the list of frequencies in the 12 tone equal temperment chromatic scale
      break;

    case 1:   //Major
      i = map(rawVal, 0, 1023, 0, 24); //24 gives three octaves of a diatonic scale
      i = i + map(degree, 0, 1023, 0, 8);
      octaveOffset = octave + (i / 8); //dividing by 8 here because a major scale only has 8notes to the octave
      i = i % 8; //modulo 8 here because a major scale has only 8 notes to the octave
      quality = majorMask[i];
      pitchNum = octaveMask[octaveOffset] + quality + tonicMask[root];
      note = chromaticScale[pitchNum];
      break;

    case 2:   //Natural Minor
      i = map(rawVal, 0, 1023, 0, 24);
      i = i + map(degree, 0, 1023, 0, 8);
      octaveOffset = octave + (i / 8);
      i = i % 8;
      quality = minorMask[i];
      pitchNum = octaveMask[octaveOffset] + quality + tonicMask[root];
      note = chromaticScale[pitchNum];
      break;

    case 3:   //Major Pentatonic
      i = map(rawVal, 0, 1023, 0, 16); //16 gives three octaves of a pentatonic scale
      i = i + map(degree, 0, 1023, 0, 5);
      octaveOffset = octave + (i / 5); //pentatonic scales have 5 notes to the octave
      i = i % 5;
      quality = majorPentatonicMask[i];
      pitchNum = octaveMask[octaveOffset] + quality + tonicMask[root];
      note = chromaticScale[pitchNum];
      break;

    case 4:   //Minor Pentatonic
      i = map(rawVal, 0, 1023, 0, 16);
      i = i + map(degree, 0, 1023, 0, 5);
      octaveOffset = octave + (i / 5);
      i = i % 5;
      quality = minorPentatonicMask[i];
      pitchNum = octaveMask[octaveOffset] + quality + tonicMask[root];
      note = chromaticScale[pitchNum];
      break;

    default: //if no match, play unquantized notes (currently unused - change scale mapping to output 0 - 5 to use)
      i = map(rawVal, 0, 1023, 31, 2878);
      note = i;
      break;
  }

}

//!!!CHANGE DETECTION FOR EACH KNOB

//Functions to determine if a knob has been turned since
//the Shift Button has changed state. This is used to only update values
//when knob changes. In other words, don't update variables
//unless the knob has been intentionally moved.
//
//This is slightly complicated by the fact that the ADC can report slightly
//different readings even when a knob is not moving

int changeRead(int buttonPin, int analogPin) {
  //when "changingLOW" is true, the shifted parameter knob will be read and updated like normal
  static bool changingLOW = false; //assume the button is not pressed at start
  //when "changingHIGH" is true, the unshifted parameter will be updated like normal
  static bool changingHIGH = true; //start with un-shifted parameter live

  //check to see if shift button is pressed.
  //if not, make our "changingLOW" flag false so no changes will happen
  //to the "shifted" parameter
  if (digitalRead(buttonPin) != LOW) {
    changingLOW = false;
  }

  //if the button is pressed, and the the changing flag was false, check
  //to see if the knob is moving by taking a few of readings and chacking to
  //see if enough of them show movement to suggest that the knob has been
  //intentionally turned. If so, set the changing flag to "true" to pass
  //all future turns of the knob.
  else if (digitalRead(buttonPin) == LOW && changingLOW == false) {
    const int READING_NUM = 8;
    int reading[READING_NUM];
    int change = 0;
    int totalChange = 0;

    for (int i = 0; i < READING_NUM; i++) {
      reading[i] = analogRead(analogPin);
      change = reading[i - 1] - reading[i]; //subtract the current reading from the former reading to record the change

      if (i > 0) { //start counting at 1 instead of zero to get rid of first reading, since the difference won't make sense
        totalChange = abs(change) + totalChange; //add up four readings, making all changes positive first
        delayMicroseconds(300);  //stabalize this reading
      }
    }

    if  (totalChange > READING_NUM - 1) { //if there is a change on every reading, assume the knob has moved
      changingLOW = true; //set a flag to tell the program to go ahead and read the konb now
      //a changingLOW flag of true will keep the above code from running every time
      return 1; //return a value of one so when we check this function we will know if we should update the "shifted" parameter
    }
  }

  else if (changingLOW == true) {
    return 1; //return a value of one here too, so when we check this function we will know if we should update the "shifted" parameter
  }

  //check to see if shift button is unpressed.
  //if not, make our "changingHIGH" flag false so no changes will happen
  if (digitalRead(buttonPin) != HIGH) {
    changingHIGH = false;
  }
  //if the shift button is not pressed and we are not currently updating the non-shifted parameter
  //we want to check to see if the knob is being intentially moved
  else if (digitalRead(buttonPin) == HIGH && changingHIGH == false) {
    const int READING_NUM = 8;
    int reading[READING_NUM];
    int change = 0;
    int totalChange = 0;

    for (int i = 0; i < READING_NUM; i++) {
      reading[i] = analogRead(analogPin);
      change = reading[i - 1] - reading[i]; //subtract the current reading from the former reading to record the change

      if (i > 0) { //start counting at 1 instead of zero to get rid of first reading, since the difference won't make sense
        totalChange = abs(change) + totalChange; //add up four readings, making all changes positive first
        delayMicroseconds(300);  //stabalize this reading
      }
    }
    if  (totalChange > READING_NUM - 1) { //if there is a change on every reading, assume the knob has moved
      changingHIGH = true; //set a flag to tell the program to go ahead and read the konb now
      return 2; //when we check this function, seeing "2" tells us to go ahaead and update the non-shifted parameter
    }
  }

  else if (changingHIGH == true) {
    return 2; //return this continuously if we are ready to update the non-shifted parameter
  }
}


int changeReadTwo(int buttonPin, int analogPin) {
  static bool changingLOW = false;
  static bool changingHIGH = true;

  //check to see if shift button is pressed.
  //if not, make our "changingLOW" flag false so no changes will happen
  //to the "shifted" parameter
  if (digitalRead(buttonPin) != LOW) {
    changingLOW = false;
  }

  //if the button is pressed, and the the changing flag was false, check
  //to see if the knob is moving by taking a few of readings and chacking to
  //see if enough of them show movement to suggest that the knob has been
  //intentionally turned. If so, set the changing flag to "true" to pass
  //all future turns of the knob.
  else if (digitalRead(buttonPin) == LOW && changingLOW == false) {
    const int READING_NUM = 8;
    int reading[READING_NUM];
    int change = 0;
    int totalChange = 0;

    for (int i = 0; i < READING_NUM; i++) {
      reading[i] = analogRead(analogPin);
      change = reading[i - 1] - reading[i]; //subtract the current reading from the former reading to record the change

      if (i > 0) { //start counting at 1 instead of zero to get rid of first reading, since the difference won't make sense
        totalChange = abs(change) + totalChange; //add up four readings, making all changes positive first
        delayMicroseconds(300);  //stabalize this reading
      }
    }

    if  (totalChange > READING_NUM - 1) { //if there is a change on every reading, assume the knob has moved
      changingLOW = true; //set a flag to tell the program to go ahead and read the konb now
      //a changingLOW flag of true will keep the above code from running every time
      return 1; //return a value of one so when we check this function we will know if we should update the "shifted" parameter
    }
  }

  else if (changingLOW == true) {
    return 1; //return a value of one here too, so when we check this function we will know if we should update the "shifted" parameter
  }

  //check to see if shift button is unpressed.
  //if not, make our "changingHIGH" flag false so no changes will happen
  if (digitalRead(buttonPin) != HIGH) {
    changingHIGH = false;
  }
  //if the shift button is not pressed and we are not currently updating the non-shifted parameter
  //we want to check to see if the knob is being intentially moved
  else if (digitalRead(buttonPin) == HIGH && changingHIGH == false) {
    const int READING_NUM = 8;
    int reading[READING_NUM];
    int change = 0;
    int totalChange = 0;

    for (int i = 0; i < READING_NUM; i++) {
      reading[i] = analogRead(analogPin);
      change = reading[i - 1] - reading[i]; //subtract the current reading from the former reading to record the change

      if (i > 0) { //start counting at 1 instead of zero to get rid of first reading, since the difference won't make sense
        totalChange = abs(change) + totalChange; //add up four readings, making all changes positive first
        delayMicroseconds(300);  //stabalize this reading
      }
    }
    if  (totalChange > READING_NUM - 1) { //if there is a change on every reading, assume the knob has moved
      changingHIGH = true; //set a flag to tell the program to go ahead and read the konb now
      return 2; //when we check this function, seeing "2" tells us to go ahaead and update the non-shifted parameter
    }
  }

  else if (changingHIGH == true) {
    return 2; //return this continuously if we are ready to update the non-shifted parameter
  }
}
